/*****************************************************************
  NAME: gopher.js
  PATH: /lib/gopher.js
  WHAT: Facilitates data transactions between this module and
        the Node Oracle Database driver "oracledb".  Returns
        a dataset (in json or as an array) based on set inputs.
******************************************************************/
var oracledb = require('oracledb');

//Set Defaults---------------------------------------
//These can be overidden by including/setting the "maxRowsReturned" or "outputFormat" properties within the "pouch" object

//NUMBER OF ROWS RETURNED:
var maxRowsReturnedDEFAULT = 2000;  //set the default for the number of rows that are be returned from queries

//CALLBACK DATA FORMAT:
var outputFormatDEFAULT = "json"; //set the default for the format used for data returned from queries. choose "json" or "array"



function Gopher (pouch) {

 //Define variables----------------------------------

    this.errorMsg = undefined;
    this.result = "ERROR";
    this.pouch = pouch;
    //these are mandatory variables that must be explicitly defined/set upon instantiation and included within the "pouch"
    this.usr = undefined;
    this.pwd = undefined;
    this.conStr = undefined;
    this.dbStmt = undefined;

    //these are mandatory, but if not explicitly defined, will be set to a default value upon instantiation
    this.outputFrmt = outputFormatDEFAULT; // use default (above) if not explicitly set/defined in the "pouch"
    this.maxRowsRetrnd = maxRowsReturnedDEFAULT; // use default (above) if not explicitly set/defined in the "pouch"

    //these variables are not mandatory
    this.bindVars = {};

 //Loop through the "pouch" object---------------------------
    for (var key in pouch) {

      if (pouch.hasOwnProperty(key)) {

          //set what keys/properties to look for, cache (from within the "pouch" object), and/or validate
          if (key === "user") {this.usr = pouch[key];} // sets the user/schema name used for logging on
          if (key === "password") {this.pwd = pouch[key];} // sets the password used for logging on
          if (key === "connectString") {this.conStr = pouch[key];} // sets the connection string used to connect to an oracle database... use the following format: host.com:port/db.com
          if (key === "dbStatement") {this.dbStmt = pouch[key];} // sets the statement/query to be executed within the database
          if (key === "outputFormat") {this.outputFrmt = pouch[key];} // sets the format the data will be returned in... including/setting this element in the "pouch" object will overide the default
          if (key === "maxRowsReturned") {this.maxRowsRetrnd = pouch[key];} // sets the max number of rows that will be returned... including/setting this element in "pouch" object will overide the default
          if (key === "bindVariables") {this.bindVars = pouch[key];} // sets the bind variables used in SQL statement... optional, use when necessary

      }
    }
}

Gopher.prototype = {

    checkPouch: function (callback) {
        //Validate inputs---------------------------------
           //these inputs must NOT be undefined
           if (this.usr  === undefined ||
               this.pwd  === undefined ||
               this.conStr  === undefined ||
               this.dbStmt  === undefined ||
               this.outputFrmt === undefined ||
               this.maxRowsRetrnd === undefined) {

                   this.errorMsg = 'INPUT INVALID: user, password, connectString, and dbStatement must NOT be \"undefined\"';
                   return callback(this.errorMsg,this.result);
           }

           //these inputs must NOT be null
           if (this.usr  === null ||
               this.pwd  === null ||
               this.conStr  === null ||
               this.dbStmt  === null ||
               this.outputFrmt === null ||
               this.maxRowsRetrnd === null) {

                   this.errorMsg = 'INPUT INVALID: user, password, connectString, and dbStatement must NOT be \"null\"';
                   return callback(this.errorMsg,this.result);
           }

           //these inputs must be a string
           if (typeof this.usr  !== "string" ||
               typeof this.pwd  !== "string" ||
               typeof this.conStr  !== "string" ||
               typeof this.dbStmt  !== "string" ||
               typeof this.outputFrmt !== "string") {

                   this.errorMsg = 'INPUT INVALID: user, password, connectString, and dbStatement must be a string';
                   return callback(this.errorMsg,this.result);
           }

           //these inputs must NOT be an empty string (blank)
           if (this.usr  === "" ||
               this.pwd  === "" ||
               this.conStr  === "" ||
               this.dbStmt  === "" ||
               this.outputFrmt  === "") {

                   this.errorMsg = 'INPUT INVALID: user, password, connectString, and dbStatement must NOT be an empty string';
                   return callback(this.errorMsg,this.result);
           }

           //maxRowsReturned must NOT be a string and must be a natual number (i.e. a positive whole number)
           var maxRowsRetrndDATATYPE = typeof this.maxRowsRetrnd;

           if (maxRowsRetrndDATATYPE !== "number" ) {

               this.errorMsg = 'INPUT INVALID: the value for \"maxRowsReturned\" must NOT be a string';
               return callback(this.errorMsg,this.result);
           }
           else if (maxRowsRetrndDATATYPE === "number"){
               //test with "regular expression" for natural numbers
               var naturalNumber = /^\+?(0|[1-9]\d*)$/.test(this.maxRowsRetrnd);
               if (naturalNumber === false) {

                   this.errorMsg = 'INPUT INVALID: the value for \"maxRowsReturned\" must be a natural number (or a positive integer)';
                   return callback(this.errorMsg,this.result);
               }
           }

           //outputFormat must equal "json" or "array"... if undefined, use default (see further below)
           if (this.outputFrmt === "json" ||
               this.outputFrmt === "array" ||
               this.outputFrmt === "" ||
               this.outputFrmt === null ||
               this.outputFrmt === undefined) {

                   //do nothing
           }
           else
               {
                   this.errorMsg = 'INPUT INVALID: unfamiliar outputFormat specified';
                   return callback(this.errorMsg,this.result);
           }

           //if the following values are used for outputFormat... use the default instead
           if (this.outputFrmt === "" ||
               this.outputFrmt === null ||
               this.outputFrmt === undefined) {

                   this.outputFrmt = outputFormatDEFAULT;
           }

           //if defined in the "pouch" Object, "bindVariables" must be stored as an array
           if (this.bindVars  !== null && this.bindVars  !== undefined && Array.isArray(this.bindVars) === true) {

                   //this.errorMsg = 'INPUT INVALID: bindVariables must be an Object';
                   return callback('INPUT INVALID: bindVariables must be an Object',this.result);
           }

           this.result = 'ALL INPUTS VALID';
           return callback(this.errorMsg,this.result);
    },

     testTunnel: function (callback) {
        var testTunnelObj = {
                "user"          : this.usr,
                "password"      : this.pwd,
                "connectString" : this.conStr,
                "dbStatement"   : "SELECT * FROM DUAL"
            };
        var gopher = new Gopher(testTunnelObj);
        //validate input values
        gopher.checkPouch(function(err,res){
            //if found invalid, by way of checkPouch (above), throw error
            if(err){return callback(err,res);}
        });
        var oracleConnectionObject = {
            "user"          : testTunnelObj.user,
            "password"      : testTunnelObj.password,
            "connectString" : testTunnelObj.connectString
        };
        var errMsg;
        oracledb.getConnection(
          oracleConnectionObject,
          function(err, connection) {
            if (err) {
                return callback(err.message,"TUNNEL FAILED");
            }
            connection.execute(
              testTunnelObj.dbStatement,
              [],
              function(err, result) {
                if (err) {
                    connection.release(function(err){
                        if (err) {
                            return callback(err.message,"TUNNEL FAILED");
                        }
                    });
                    return callback(err.message,"TUNNEL FAILED");
                }
                connection.release(function(err){
                    if (err) {
                        return callback(err.message,"TUNNEL FAILED");
                    }
                });
                if (result.rows == 'X') {
                    resMsg = "TUNNEL USABLE";
                }
                else {
                    resMsg = "TUNNEL FAILED";
                }
                return callback(undefined,resMsg);
            });
          });
    },

    releaseTunnel: function (connection) {

          connection.release(
            function(err) {
              if (err) {
                console.error(err.message);
              }
            });

    },

    tunnel: function (callback) {
        var gopher = new Gopher(this.pouch);
        //validate input values
        gopher.checkPouch(function(err,res){
            //if found invalid, by way of checkPouch (above), throw error
            if(err){return callback(err,res);}
        });

        //configure inputs for use with oracle driver
        var oracleConnectionObject = {
            "user"          : this.usr,
            "password"      : this.pwd,
            "connectString" : this.conStr
            };
        var dbStatement = this.dbStmt;
        var transactionConfig = {
            "maxRows"       : this.maxRowsRetrnd
            };
        var bindVars = this.bindVars;
        var format = this.outputFrmt;

        oracledb.getConnection(
          oracleConnectionObject ,
          function(err,connection) {
            if (format === 'array'){transactionConfig.outFormat = oracledb.ARRAY;}
            if (format === 'json'){transactionConfig.outFormat = oracledb.OBJECT;}
            if (err) {
              return callback(err.message,'TUNNEL ERROR');
            }
            connection.execute(
              dbStatement,
              bindVars,
              transactionConfig,
              function (err, res) {
                if (err) {
                  gopher.releaseTunnel(connection); //close connection
                  return callback(err.message,'TUNNEL ERROR');
                }
                gopher.releaseTunnel(connection); //close connection
                return callback(undefined,res.rows);
              }
          );
          }
      );
    }
};

module.exports = Gopher;
